@@ -1517,7 +1517,7 @@ tegra_nand_probe(struct platform_device *pdev)
 #ifdef CONFIG_MTD_PARTITIONS
 	err = parse_mtd_partitions(mtd, part_probes, &info->parts, 0);
 	if (err > 0) {
-		err = add_mtd_partitions(mtd, info->parts, err);
+		add_mtd_partitions(mtd, info->parts, err);
 	} else if (err <= 0 && plat->parts) {
 		err = add_mtd_partitions(mtd, plat->parts, plat->nr_parts);
 	} else
