@@ -158,15 +158,15 @@ static int qcnet_bind(struct usbnet *usbnet, struct usb_interface *iface)
 	int dir_in, dir_out, xfer_int;
 
 	if (iface->num_altsetting != 1) {
-		GOBI_ERROR("invalid num_altsetting %u", iface->num_altsetting);
-		return -EINVAL;
+		GOBI_WARN("invalid num_altsetting %u", iface->num_altsetting);
+		return -ENODEV;
 	}
 
 	if (iface->cur_altsetting->desc.bInterfaceNumber != 0
 	    && iface->cur_altsetting->desc.bInterfaceNumber != 5) {
-		GOBI_ERROR("invalid interface %d",
+		GOBI_WARN("invalid interface %d",
 			  iface->cur_altsetting->desc.bInterfaceNumber);
-		return -EINVAL;
+		return -ENODEV;
 	}
 
 	GOBI_DEBUG("interface number: %d",
@@ -187,15 +187,15 @@ static int qcnet_bind(struct usbnet *usbnet, struct usb_interface *iface)
 
 		if (dir_in && !xfer_int) {
 			if (in) {
-				GOBI_ERROR("multiple in endpoints");
-				return -EINVAL;
+				GOBI_WARN("multiple in endpoints");
+				return -ENODEV;
 			}
 			GOBI_DEBUG("setting endpoint %d as in", i);
 			in = endpoint;
 		} else if (dir_out && !xfer_int) {
 			if (out) {
-				GOBI_ERROR("multiple out endpoints");
-				return -EINVAL;
+				GOBI_WARN("multiple out endpoints");
+				return -ENODEV;
 			}
 			GOBI_DEBUG("setting endpoint %d as out", i);
 			out = endpoint;
@@ -205,23 +205,23 @@ static int qcnet_bind(struct usbnet *usbnet, struct usb_interface *iface)
 	}
 
 	if (!in || !out) {
-		GOBI_ERROR("missing endpoint(s)");
+		GOBI_WARN("missing endpoint(s)");
 		if (in)
-			GOBI_ERROR("found in endpoint: %u",
+			GOBI_WARN("found in endpoint: %u",
 				in->desc.bEndpointAddress);
 		else
-			GOBI_ERROR("didn't find in endpoint");
+			GOBI_WARN("didn't find in endpoint");
 		if (out)
-			GOBI_ERROR("found out endpoint: %u",
+			GOBI_WARN("found out endpoint: %u",
 				out->desc.bEndpointAddress);
 		else
-			GOBI_ERROR("didn't find out endpoint");
-		return -EINVAL;
+			GOBI_WARN("didn't find out endpoint");
+		return -ENODEV;
 	}
 
 	if (usb_set_interface(usbnet->udev,
 			      iface->cur_altsetting->desc.bInterfaceNumber, 0))	{
-		GOBI_ERROR("unable to set interface");
+		GOBI_WARN("unable to set interface");
 		return -EINVAL;
 	}
 
@@ -243,14 +243,8 @@ static void qcnet_unbind(struct usbnet *usbnet, struct usb_interface *iface)
 
 	dev->dying = true;
 
-	iface->needs_remote_wakeup = 0;
-	netif_carrier_off(usbnet->net);
-	qc_deregister(dev);
-
 	kfree(usbnet->net->netdev_ops);
 	usbnet->net->netdev_ops = NULL;
-	/* drop the list's ref */
-	qcusbnet_put(dev);
 }
 
 static void qcnet_bg_complete(struct work_struct *work)
@@ -518,21 +512,21 @@ static int discover_endpoints(struct qcusbnet *dev)
 		if (dir_in && xfer_int) {
 			if (int_in) {
 				GOBI_ERROR("multiple int_in endpoints");
-				return -EINVAL;
+				return -ENODEV;
 			}
 			GOBI_DEBUG("setting endpoint %d as int in", i);
 			int_in = endpoint;
 		} else if (dir_in && !xfer_int) {
 			if (bulk_in) {
 				GOBI_ERROR("multiple bulk_in endpoints");
-				return -EINVAL;
+				return -ENODEV;
 			}
 			GOBI_DEBUG("setting endpoint %d as bulk in", i);
 			bulk_in = endpoint;
 		} else if (dir_out && !xfer_int) {
 			if (bulk_out) {
 				GOBI_ERROR("multiple bulk_out endpoints");
-				return -EINVAL;
+				return -ENODEV;
 			}
 			GOBI_DEBUG("setting endpoint %d as bulk out", i);
 			bulk_out = endpoint;
@@ -558,7 +552,7 @@ static int discover_endpoints(struct qcusbnet *dev)
 				bulk_out->desc.bEndpointAddress);
 		else
 			GOBI_ERROR("didn't find bulk_out endpoint");
-		return -EINVAL;
+		return -ENODEV;
 	}
 
 	dev->iface_num     = dev->iface->cur_altsetting->desc.bInterfaceNumber;
@@ -586,25 +580,28 @@ int qcnet_probe(struct usb_interface *iface, const struct usb_device_id *vidpids
 	status = usbnet_probe(iface, vidpids);
 	if (status < 0) {
 		GOBI_WARN("usbnet_probe failed: %d", status);
-		return status;
+		goto fail;
 	}
 
 	usbnet = usb_get_intfdata(iface);
 
 	if (!usbnet) {
 		GOBI_ERROR("usbnet is NULL");
-		return -ENXIO;
+		status = -ENXIO;
+		goto fail_disconnect;
 	}
 
 	if (!usbnet->net) {
 		GOBI_ERROR("usbnet->net is NULL");
-		return -ENXIO;
+		status = -ENXIO;
+		goto fail_disconnect;
 	}
 
 	dev = kmalloc(sizeof(struct qcusbnet), GFP_KERNEL);
 	if (!dev) {
 		GOBI_ERROR("failed to allocate struct qcusbnet");
-		return -ENOMEM;
+		status = -ENOMEM;
+		goto fail_disconnect;
 	}
 
 	dev->dying = false;
@@ -613,8 +610,7 @@ int qcnet_probe(struct usb_interface *iface, const struct usb_device_id *vidpids
 	status = discover_endpoints(dev);
 	if (status) {
 		GOBI_ERROR("discover_endpoints failed: %d", status);
-		kfree(dev);
-		return status;
+		goto fail_free;
 	}
 
 	usbnet->data[0] = (unsigned long)dev;
@@ -624,7 +620,8 @@ int qcnet_probe(struct usb_interface *iface, const struct usb_device_id *vidpids
 	netdevops = kmalloc(sizeof(struct net_device_ops), GFP_KERNEL);
 	if (!netdevops) {
 		GOBI_ERROR("failed to allocate net device ops");
-		return -ENOMEM;
+		status = -ENOMEM;
+		goto fail_free;
 	}
 	memcpy(netdevops, usbnet->net->netdev_ops, sizeof(struct net_device_ops));
 
@@ -667,15 +664,17 @@ int qcnet_probe(struct usb_interface *iface, const struct usb_device_id *vidpids
 
 	status = qc_register(dev);
 	if (status) {
-		qc_deregister(dev);
-	} else {
-		iface->needs_remote_wakeup = 1;
-
-		mutex_lock(&qcusbnet_lock);
-		/* Give our initial ref to the list */
-		list_add(&dev->node, &qcusbnet_list);
-		mutex_unlock(&qcusbnet_lock);
+		GOBI_ERROR("qc_register failed: %d", status);
+		goto fail_destroy;
 	}
+
+	iface->needs_remote_wakeup = 1;
+
+	mutex_lock(&qcusbnet_lock);
+	/* Give our initial ref to the list */
+	list_add(&dev->node, &qcusbnet_list);
+	mutex_unlock(&qcusbnet_lock);
+
 	/* After calling qc_register, MEID is valid */
 	addr = &usbnet->net->dev_addr[0];
 	for (i = 0; i < 6; i++)
@@ -684,6 +683,15 @@ int qcnet_probe(struct usb_interface *iface, const struct usb_device_id *vidpids
 	addr[0] &= 0xfe;		/* clear multicast bit */
 	addr[0] |= 0x02;		/* set local assignment bit (IEEE802) */
 
+	return 0;
+
+fail_destroy:
+	destroy_workqueue(dev->workqueue);
+fail_free:
+	kfree(dev);
+fail_disconnect:
+	usbnet_disconnect(iface);
+fail:
 	return status;
 }
 EXPORT_SYMBOL_GPL(qcnet_probe);
@@ -694,13 +702,20 @@ static void qcnet_disconnect(struct usb_interface *intf)
 	struct qcusbnet *dev = (struct qcusbnet *)usbnet->data[0];
 	struct list_head *node, *tmp;
 	struct urb *urb;
+
+	intf->needs_remote_wakeup = 0;
+	netif_carrier_off(usbnet->net);
+	usbnet_disconnect(intf);
+
+	qc_deregister(dev);
+
 	destroy_workqueue(dev->workqueue);
 	list_for_each_safe(node, tmp, &dev->urbs) {
 		urb = list_entry(node, struct urb, urb_list);
 		list_del(&urb->urb_list);
 		free_urb_with_skb(urb);
 	}
-	usbnet_disconnect(intf);
+	qcusbnet_put(dev);
 }
 
 static struct usb_driver qcusbnet = {
